/* Copyright (C) 2023-2023 Warsaw University of Technology, Warsaw
    SPDX-License-Identifier: GPL-3.0-only
    Authors: Daniel Wielanek [committer] */
#ifndef NICACBMATCONTAINER_H_
#define NICACBMATCONTAINER_H_


#include <TNamed.h>

#include <AnalysisTree/Detector.hpp>
#include <AnalysisTree/EventHeader.hpp>
#include <AnalysisTree/Matching.hpp>

class TChain;
class TFile;
//#define DEBUG_EXTR

namespace AnalysisTree
{
  class Configuration;
}
namespace Hal
{
  class InputDataInfo;
}

class HalCbmATIOManager;
struct AnaTreeRecoIds {
  Int_t vtx_px, vtx_py, vtx_pz;
  Int_t vtx_dcax, vtx_dcay, vtx_dcaz;
  Int_t vtx_vtxchi2, vtx_chi2;
  Int_t vtx_q, vtx_nhits, vtx_mvdhits;
  Int_t vtx_x, vtx_cx0, vtx_cov1;
  Int_t tof_mass2;
};

class CbmAnaTreeContainer : public TNamed {

 public:
  CbmAnaTreeContainer() : TNamed(){};
  virtual void LoadFields(std::vector<TString> list) = 0;
  virtual Bool_t ConnectToTree(TChain* chains)       = 0;
  AnalysisTree::BranchConfig GetBranchConf(std::vector<TString> files, TString patternToFind, Bool_t& found) const;
  virtual ~CbmAnaTreeContainer(){};
  ClassDef(CbmAnaTreeContainer, 1)
};


class CbmAnaTreeRecoContainer : public CbmAnaTreeContainer {
  AnalysisTree::EventHeader* fEvent       = {nullptr};
  AnalysisTree::TrackDetector* fVtxTracks = {nullptr};
  AnalysisTree::Matching* fVtx2Tof        = {nullptr};
  AnalysisTree::HitDetector* fTofHits     = {nullptr};
  AnalysisTree::Matching* fVtx2Mc         = {nullptr};
  AnaTreeRecoIds fVtxIds;

 public:
  CbmAnaTreeRecoContainer(){};
  void LoadFields(std::vector<TString> list);
  Bool_t ConnectToTree(TChain* chains);
  inline AnalysisTree::EventHeader* GetEventHeader() const { return fEvent; };
  inline AnalysisTree::TrackDetector* GetVtxTracks() const { return fVtxTracks; };
  inline AnalysisTree::HitDetector* GetTofHits() const { return fTofHits; };
  inline AnalysisTree::Matching* GetVtx2ToFMatch() const { return fVtx2Tof; };
  inline AnalysisTree::Matching* GetVtx2Sim() const { return fVtx2Mc; };
  inline AnaTreeRecoIds& GetFieldIds() { return fVtxIds; }
  virtual ~CbmAnaTreeRecoContainer()
  {
    if (fEvent) {
      delete fEvent;
      delete fVtxTracks;
      delete fTofHits;
      delete fVtx2Tof;
      delete fVtx2Mc;
    }
  }
  ClassDef(CbmAnaTreeRecoContainer, 1)
};

struct AnaTreeMcIds {
  Int_t px, py, pz;
  Int_t pdg, motherId, mass;
  Int_t event_b;
  Int_t event_psi;
  Int_t freezX, freezY, freezZ, freezT;
};

class CbmAnaTreeMcContainer : public CbmAnaTreeContainer {
  AnalysisTree::EventHeader* fEvent   = {nullptr};
  AnalysisTree::Particles* fParticles = {nullptr};
  AnaTreeMcIds fIds;

 public:
  CbmAnaTreeMcContainer(){};
  void LoadFields(std::vector<TString> list);
  Bool_t ConnectToTree(TChain* chains);
  inline AnaTreeMcIds& GetFieldIds() { return fIds; };
  inline AnalysisTree::EventHeader* GetEventHeader() const { return fEvent; };
  inline AnalysisTree::Particles* GetParticles() const { return fParticles; };
  virtual ~CbmAnaTreeMcContainer()
  {
    if (fEvent) {
      delete fEvent;
      delete fParticles;
    }
  };
  ClassDef(CbmAnaTreeMcContainer, 1)
};

struct AnaTreeV0Ids {
  Int_t chi2geo, chi2_prim_first;
  Int_t chi2_prim_second, chi2_topo;
  Int_t cosine_first, cosine_second, cosine_topo;
  Int_t distance, l, l_over_dl;
  Int_t px, py, pz;
  Int_t x, y, z;
  Int_t dau1id, dau2id, id;
  Int_t pdg, mass;
#ifdef DEBUG_EXTR
  int dau1px, dau1py, dau1pz, dau1x, dau1y, dau1z;
  int dau2px, dau2py, dau2pz, dau2x, dau2y, dau2z;
#endif
};

class CbmAnaTreeV0Container : public CbmAnaTreeContainer {
  AnalysisTree::Particles* fParticles   = {nullptr};
  AnalysisTree::Particles* fParticlesMc = {nullptr};
  AnalysisTree::Matching* fV02Mc        = {nullptr};
  AnaTreeV0Ids fIds;

 public:
  CbmAnaTreeV0Container(){};
  void LoadFields(std::vector<TString> list);
  Bool_t ConnectToTree(TChain* chains);
  inline AnaTreeV0Ids& GetFieldIds() { return fIds; };
  inline AnalysisTree::Particles* GetParticles() const { return fParticles; };
  AnalysisTree::Matching* GetMaching() const { return fV02Mc; };
  virtual ~CbmAnaTreeV0Container()
  {
    if (fParticles) delete fParticles;
    if (fParticlesMc) delete fParticlesMc;
    if (fV02Mc) delete fV02Mc;
  };
  ClassDef(CbmAnaTreeV0Container, 1)
};


class CbmAnaTreeContainers : public TObject {


 public:
  enum class EContainerType
  {
    kMc   = 0,
    kReco = 1,
    kV0   = 2
  };
  CbmAnaTreeContainer* fContainers[3] = {nullptr, nullptr, nullptr};
  CbmAnaTreeContainers();
  void LoadConf(Hal::InputDataInfo* info);
  static std::vector<TString> GetNames();
  CbmAnaTreeContainer* GetContainer(EContainerType type);
  CbmAnaTreeV0Container* GetV0Data() const { return (CbmAnaTreeV0Container*) fContainers[2]; }
  CbmAnaTreeRecoContainer* GetRecoData() const { return (CbmAnaTreeRecoContainer*) fContainers[2]; }
  CbmAnaTreeMcContainer* GetSimData() const { return (CbmAnaTreeMcContainer*) fContainers[2]; }
  Int_t GetSize() const { return 3; }
  virtual ~CbmAnaTreeContainers(){};
  ClassDef(CbmAnaTreeContainers, 1)
};


#endif /* NICACBMATCONTAINER_H_ */
