/* Copyright (C) 2025 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergei Zharko [committer] */

/// \file   CbmTaskInspectRecoTimeslice.cxx
/// \brief  A task to store reconstructed data into a ROOT tree (implementation)
/// \author Sergei Zharko <s.zharko@gsi.de>
/// \since  26.03.2025

#include "CbmTaskInspectRecoTimeslice.h"

#include "FairRootManager.h"

#include <type_traits>

// ---------------------------------------------------------------------------------------------------------------------
//
CbmTaskInspectRecoTimeslice::CbmTaskInspectRecoTimeslice() : FairTask("InspectOnlineRecoTimeslice") {}


// ---------------------------------------------------------------------------------------------------------------------
//
void CbmTaskInspectRecoTimeslice::Exec(Option_t*)
{
  LOG(info) << GetName() << ": timeslice " << fNumTs << " with " << fBmonHits->NElements() << " BMON hits, "
            << fStsHits->NElements() << " STS hits, " << fTrdHits->NElements() << " TRD hits, " << fTofHits->NElements()
            << " TOF hits, " << fTracks->size() << " tracks";

  ++fNumTs;
}


// ---------------------------------------------------------------------------------------------------------------------
//
void CbmTaskInspectRecoTimeslice::Finish()
{
  LOG(info) << "=====================================";
  LOG(info) << GetName() << ": Run summary";
  LOG(info) << "Timeslices : " << fNumTs;
  LOG(info) << "=====================================";
}


// ---------------------------------------------------------------------------------------------------------------------
//
InitStatus CbmTaskInspectRecoTimeslice::Init()
{
  auto* ioman = FairRootManager::Instance();
  if (!ioman) {
    LOG(fatal) << "CbmSourceRecoTimeslice::Init(): FairRootManager is not defined";
    return kFATAL;
  }

  LOG(info) << "==================================================";
  LOG(info) << GetName() << ": Initialising...";

  auto InitBranch = [&](const auto*& vec, const char* name) -> bool {
    using VecPtr_t = std::remove_reference_t<decltype(vec)>;
    vec            = ioman->InitObjectAs<VecPtr_t>(name);
    if (!vec) {
      LOG(error) << GetName() << ": branch " << name << " not found";
      return false;
    }
    LOG(info) << GetName() << ": found branch " << name << " at " << vec;
    return true;
  };

  InitBranch(fBmonHits, "OnlineBmonHit");
  InitBranch(fStsHits, "OnlineStsHit");
  InitBranch(fTrdHits, "OnlineTrdHit");
  InitBranch(fTofHits, "OnlineTofHit");
  InitBranch(fTracks, "OnlineTrack");
  InitBranch(fTrackStsHitIndices, "OnlineTrackStsHitIndex");
  InitBranch(fTrackTrdHitIndices, "OnlineTrackTrdHitIndex");
  InitBranch(fTrackTofHitIndices, "OnlineTrackTofHitIndex");

  return kSUCCESS;
}
